package geniusweb.protocol.tournament.allpermutations;

import java.net.URI;
import java.net.URISyntaxException;
import java.util.Collections;
import java.util.List;

import com.fasterxml.jackson.annotation.JsonCreator;
import com.fasterxml.jackson.annotation.JsonProperty;

import geniusweb.profile.Profile;
import geniusweb.protocol.session.SessionSettings;
import geniusweb.protocol.session.TeamOfPartiesAndProfiles;
import geniusweb.protocol.session.saop.SAOPSettings;
import geniusweb.protocol.session.saop.SaopPartyWithProfile;
import geniusweb.protocol.session.shaop.SHAOPSettings;
import geniusweb.protocol.session.shaop.ShaopTeam;
import geniusweb.protocol.tournament.TournamentProtocol;
import geniusweb.protocol.tournament.TournamentSettings;
import geniusweb.references.Parameters;
import geniusweb.references.PartyRef;
import geniusweb.references.PartyWithParameters;
import geniusweb.references.PartyWithProfile;
import geniusweb.references.ProfileRef;
import tudelft.utilities.immutablelist.FixedList;
import tudelft.utilities.immutablelist.Function2;
import tudelft.utilities.immutablelist.ImmutableList;
import tudelft.utilities.immutablelist.MapList;
import tudelft.utilities.immutablelist.MapThreadList;
import tudelft.utilities.immutablelist.Permutations;
import tudelft.utilities.immutablelist.PermutationsOrderedWithoutReturn;
import tudelft.utilities.immutablelist.PermutationsWithReturn;
import tudelft.utilities.immutablelist.PermutationsWithoutReturn;
import tudelft.utilities.immutablelist.Tuple;
import tudelft.utilities.immutablelist.Tuples;
import tudelft.utilities.logging.Reporter;

/**
 * Takes a number of parties and profiles and makes all permutations of them
 * with the given number of parties and profiles. All profiles must be in the
 * same domain.
 * <p>
 * All sessions are based on the given {@link SessionSettings} (can be
 * {@link SAOPSettings} or {@link SHAOPSettings}). The parties for each
 * tournament session are added to this basic sessionsettings. sessionsettings
 * contains basically all the normal run-sessionsettings. This is used as a
 * "template" for all sessions of the tournament. You can put any use any
 * session setting here, and each session will be run according to the protocol
 * you select here.
 * 
 * The participants list usually is empty. The AllPermutationsProtocol adds the
 * the required parties to this list. So if you provide a non-empty list here,
 * then these parties would be present in every session in the tournament.
 * <p>
 * If the tournament is used with a SHAOP sessionsettings, a number of rules are
 * used to generate the COB party from the settings.
 * 
 * The COB party is assumed to be the comparebids-1.0.0 party on the same server
 * as the party you provide. So if you provide http://server/../party1-1.2.3
 * then the cob party will be assumed to be at
 * http://server/../comparebids-1.0.0. The provided parties that you provide are
 * the SHAOP parties (you can also use SAOP parties). The provided profile is
 * also given to the SHAOP party. Be aware that you typically need a parameter
 * like partial=10. The COB party gets the profile as specified, but without the
 * query part. The COB and SHAOP party are thus a fixed pair.
 */
public class AllPermutationsSettings implements TournamentSettings {

	private final List<PartyWithParameters> parties;
	private final List<ProfileRef> profiles;
	private final boolean reuseParties;
	private final int partiesPerSession;
	private final SessionSettings sessionsettings;

	/**
	 * 
	 * @param parties           a list of {@link PartyRef}s. Must contain at
	 *                          least <partiesPerSession> elements.
	 * @param reuseParties      if true, we use PermutationsWithReturn, if false
	 *                          we use PermutationsWithoutReturn to create the
	 *                          parties.
	 * @param profiles          list of available {@link Profile}s to be
	 *                          permutated over the parties
	 * @param partiesPerSession number of parties per session, must be at least
	 *                          2.
	 * @param sesettings        The generic {@link SessionSetting}.
	 *                          {@link SessionSettings#with(PartyWithProfile)}
	 *                          will be used to add the required
	 *                          {@link PartyWithProfile}s
	 */
	@JsonCreator
	public AllPermutationsSettings(
			@JsonProperty("parties") List<PartyWithParameters> parties,
			@JsonProperty("reuseParties") boolean reuseParties,
			@JsonProperty("profiles") List<ProfileRef> profiles,
			@JsonProperty("partiesPerSession") int partiesPerSession,
			@JsonProperty("sessionsettings") SessionSettings sesettings) {
		if (partiesPerSession < 2)
			throw new IllegalArgumentException("partiesPerSession must be >=2");
		if (parties == null || parties.size() < partiesPerSession)
			throw new IllegalArgumentException("parties must contain at least "
					+ partiesPerSession + " PartyRefs");
		if (profiles == null || profiles.size() < 2)
			throw new IllegalArgumentException("profiles must contain at least "
					+ partiesPerSession + " ProfileRefs");
		this.partiesPerSession = partiesPerSession;
		this.reuseParties = reuseParties;
		this.parties = parties;
		this.profiles = profiles;
		this.sessionsettings = sesettings;
	}

	@Override
	public Double getMaxRunTime() {
		return permutations().size().doubleValue()
				* (sessionsettings.getMaxRunTime() + 2d);
	}

	@Override
	public TournamentProtocol getProtocol(Reporter logger) {
		AllPermutationsState state = new AllPermutationsState(this,
				permutations(), Collections.emptyList());
		return new AllPermutationsProtocol(state, logger);
	}

	/**
	 * @return {@link ImmutableList} containing all {@link SessionSettings} for
	 *         all Sessions in the correct order.
	 */
	public ImmutableList<SessionSettings> permutations() {
		ImmutableList<ImmutableList<TeamOfPartiesAndProfiles>> partylistlist;
		ImmutableList<PartyWithParameters> partieslist = new FixedList<PartyWithParameters>(
				parties);
		ImmutableList<ProfileRef> profileslist = new FixedList<ProfileRef>(
				profiles);
		partylistlist = getParticipants(partieslist, profileslist,
				partiesPerSession, reuseParties);
		return new MapList<>(partyproflist -> createSetting(partyproflist),
				partylistlist);
	}

	@Override
	public String toString() {
		return "AllPermutationsSettings[" + parties + "," + reuseParties + ","
				+ profiles + "," + partiesPerSession + "," + sessionsettings
				+ "]";
	}

	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result + ((parties == null) ? 0 : parties.hashCode());
		result = prime * result + partiesPerSession;
		result = prime * result
				+ ((profiles == null) ? 0 : profiles.hashCode());
		result = prime * result + (reuseParties ? 1231 : 1237);
		result = prime * result
				+ ((sessionsettings == null) ? 0 : sessionsettings.hashCode());
		return result;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		AllPermutationsSettings other = (AllPermutationsSettings) obj;
		if (parties == null) {
			if (other.parties != null)
				return false;
		} else if (!parties.equals(other.parties))
			return false;
		if (partiesPerSession != other.partiesPerSession)
			return false;
		if (profiles == null) {
			if (other.profiles != null)
				return false;
		} else if (!profiles.equals(other.profiles))
			return false;
		if (reuseParties != other.reuseParties)
			return false;
		if (sessionsettings == null) {
			if (other.sessionsettings != null)
				return false;
		} else if (!sessionsettings.equals(other.sessionsettings))
			return false;
		return true;
	}

	/**
	 * 
	 * @param partyproflist a list of {@link PartyWithProfile} settings
	 * @return {@link SessionSettings}
	 */
	private SessionSettings createSetting(
			ImmutableList<TeamOfPartiesAndProfiles> partyproflist) {
		SessionSettings settings = sessionsettings;
		for (TeamOfPartiesAndProfiles partyprof : partyproflist) {
			settings = settings.with(partyprof);
		}
		return settings;
	}

	/**
	 * 
	 * @param parties
	 * @param profiles
	 * @param n                    number of items to draw. Assumed >= 1.
	 * @param drawPartyWithPutback if parties can be drawn multiple times
	 * @return list of sessionsets where each sessionset contains settings for a
	 *         session: a list with {@link TeamOfPartiesAndProfiles}s. The
	 *         sessionsets are made by making all combinations of parties and
	 *         profiles. Profiles are drawn with replace.
	 */
	private ImmutableList<ImmutableList<TeamOfPartiesAndProfiles>> getParticipants(
			ImmutableList<PartyWithParameters> parties,
			ImmutableList<ProfileRef> profiles, int n,
			boolean drawPartyWithPutback) {

		Permutations<PartyWithParameters> partiesPermutations;
		if (drawPartyWithPutback) {
			partiesPermutations = new PermutationsWithReturn<>(parties, n);
		} else {
			partiesPermutations = new PermutationsWithoutReturn<>(parties, n);
		}

		Permutations<ProfileRef> profilesPermutations = new PermutationsOrderedWithoutReturn<>(
				profiles, n);

		// each tuple contains sesion info: a list of partyref and a list of
		// profileref's.
		Tuples<ImmutableList<PartyWithParameters>, ImmutableList<ProfileRef>> tuples = new Tuples<>(
				partiesPermutations, profilesPermutations);

		return new MapList<Tuple<ImmutableList<PartyWithParameters>, ImmutableList<ProfileRef>>, ImmutableList<TeamOfPartiesAndProfiles>>(
				tuple -> teamlist(tuple), tuples);
	}

	/**
	 * 
	 * @param tuple a tuple with (1) ImmutableList<PartyRef> (2)
	 *              ImmutableList<ProfileRef>
	 * @return list of TeamOfPartiesAndProfiles, each picked from the two lists
	 *         in order.
	 */
	private ImmutableList<TeamOfPartiesAndProfiles> teamlist(
			Tuple<ImmutableList<PartyWithParameters>, ImmutableList<ProfileRef>> tuple) {
		Function2<PartyWithParameters, ProfileRef, TeamOfPartiesAndProfiles> //
		makeparty = new Function2<PartyWithParameters, ProfileRef, TeamOfPartiesAndProfiles>() {
			@Override
			public TeamOfPartiesAndProfiles apply(PartyWithParameters party,
					ProfileRef profile) {
				if (sessionsettings instanceof SAOPSettings)
					return new SaopPartyWithProfile(party, profile);
				else if (sessionsettings instanceof SHAOPSettings) {
					return new ShaopTeam(new PartyWithProfile(party, profile),
							makeCob(party.getPartyRef(), profile));
				} else
					throw new IllegalArgumentException(
							"Unknown/unsupported protocol "
									+ sessionsettings.getClass());
			}

		};
		return new MapThreadList<TeamOfPartiesAndProfiles, PartyWithParameters, ProfileRef>(
				makeparty, tuple.get1(), tuple.get2());
	}

	/**
	 * Extract the raw URL, without the part after the question mark
	 * 
	 * @param party   a party reference (for another party)
	 * @param profile a profile that may have a filter like "?partial=XX"
	 * @return profile without the filter and with a cob party on the same
	 *         machine. IT IS ASSUMED that there will the default
	 *         "comparebids-1.0.0" party on the same machine as where the other
	 *         party is
	 */
	protected static PartyWithProfile makeCob(PartyRef party,
			ProfileRef profile) {
		try {
			URI profileuri = profile.getURI();
			URI partyuri = party.getURI();
			ProfileRef cobprof = new ProfileRef(profileuri.getScheme() + "://"
					+ profileuri.getAuthority() + profileuri.getPath());
			String partypath = partyuri.getPath();
			partypath = partypath.substring(0, partypath.lastIndexOf('/'));
			PartyWithParameters cobparty = new PartyWithParameters(new PartyRef(
					partyuri.getScheme() + "://" + partyuri.getAuthority()
							+ partypath + "/comparebids-1.0.0"),
					new Parameters());
			return new PartyWithProfile(cobparty, cobprof);

		} catch (URISyntaxException e) {
			throw new IllegalArgumentException(
					"Failed making cob party with profile " + profile, e);
		}
	}

}
